/*
Pupilles

Copyright (C) 2004 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU.

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

import simu.*;

import TFLib;

/**
 * Algorithme de simulation pour SimuApplet.
 * Figure de diffraction de pupilles de formes varies
 * compilation: javac13 -target 1.1 -encoding ISO-8859-1 -classpath .:SimuLab.jar Pupilles.java
 */
public class Pupilles extends SimuApplet {
    
	final static int dimx = 256;
	final static int dimy = 256;
	
    String pupille;
    double diametre, diametre2;
    double distance;
	double cote;
	String methode;
	String echelle;
	boolean quadrants;
	double lambda;
	boolean zoom;
	double[][] imagep;
	int etape;
	
    public void init() {
        ParamIn pupille = new ParamIn();
        pupille.type = "string";
        pupille.label = "pupille";
        pupille.titre = "pupille";
        pupille.acquisition = "choix";
        String[] fchoix = {"circulaire", "circ. + obst. second.", "2 circulaires", "croix d'Angel"};
        pupille.choix = fchoix;
        pupille.defaut = "circulaire";
        addParamIn(pupille);
        
        ParamIn diametre = new ParamIn();
        diametre.type = "nombre";
        diametre.label = "diametre";
        diametre.titre = "diamtre";
        diametre.unit = "m";
        diametre.acquisition = "champ";
        diametre.defaut = "1";
        addParamIn(diametre);
		
        ParamIn diametre2 = new ParamIn();
        diametre2.type = "nombre";
        diametre2.label = "diametre2";
        diametre2.titre = "diamtre 2";
        diametre2.unit = "m";
        diametre2.acquisition = "champ";
        diametre2.defaut = "0.5";
        addParamIn(diametre2);
		
        ParamIn distance = new ParamIn();
        distance.type = "nombre";
        distance.label = "distance";
        distance.titre = "distance";
        distance.unit = "m";
        distance.acquisition = "champ";
        distance.defaut = "4";
        addParamIn(distance);
         
        ParamIn cote = new ParamIn();
        cote.type = "nombre";
        cote.label = "cote";
        cote.titre = "taille ct";
        cote.unit = "m";
        cote.acquisition = "champ";
        cote.defaut = "8";
        addParamIn(cote);
         
        ParamIn methode = new ParamIn();
        methode.type = "string";
        methode.label = "methode";
        methode.titre = "mthode";
        methode.acquisition = "choix";
        String[] mchoix = {"FFT", "FT"};
        methode.choix = mchoix;
        methode.defaut = "FFT";
        addParamIn(methode);
        
        ParamIn echelle = new ParamIn();
        echelle.type = "string";
        echelle.label = "echelle";
        echelle.titre = "chelle";
        echelle.acquisition = "choix";
        String[] echoix = {"linaire", "logarithmique"};
        echelle.choix = echoix;
        echelle.defaut = "linaire";
        addParamIn(echelle);
        
        ParamIn quadrants = new ParamIn();
        quadrants.type = "string";
        quadrants.label = "quadrants";
        quadrants.titre = "changer les quadrants";
        quadrants.acquisition = "case";
        quadrants.defaut = "on";
        addParamIn(quadrants);
        
        ParamIn lambda = new ParamIn();
        lambda.type = "nombre";
        lambda.label = "lambda";
        lambda.titre = "lambda";
        lambda.unit = "m";
        lambda.acquisition = "champ";
        lambda.defaut = "0.5E-6";
        addParamIn(lambda);
        
        ParamIn zoom = new ParamIn();
        zoom.type = "string";
        zoom.label = "zoom";
        zoom.titre = "zoom";
        zoom.acquisition = "case";
        zoom.defaut = "off";
        addParamIn(zoom);
        
        // paramtres de sortie
        ParamOut imagep = new ParamOut();
        imagep.type = "tableau";
        imagep.label = "imagep";
        imagep.titre = "image de la pupille et figure de diffraction";
        addParamOut(imagep);
        
        ParamOut info = new ParamOut();
        info.type = "string";
        info.label = "info";
        addParamOut(info);
        
       // options d'affichage
        Affichage affimage = new Affichage();
        affimage.type = "image";
        affimage.titre = "Image de la pupille et figure de diffraction (PSF)";
        String[][] pparams1 = {{"imagep"}};
        affimage.params = pparams1;
        affimage.imgdimx = dimx*2+1;
        affimage.imgdimy = dimy;
        
        addAffichage(affimage);
        super.init();
	}


    public void initCalculLive(ListeValeurs in) throws SimuException {
        pupille = in.lireString("pupille");
        diametre = in.lireDouble("diametre");
        diametre2 = in.lireDouble("diametre2");
        distance = in.lireDouble("distance");
        cote = in.lireDouble("cote");
		methode = in.lireString("methode");
		echelle = in.lireString("echelle");
		quadrants = ("on".equals(in.lireString("quadrants")));
        lambda = in.lireDouble("lambda");
		zoom = ("on".equals(in.lireString("zoom")));
		etape = 0;
    }
    
    public ListeValeurs calculLive() throws SimuException {
		if (etape == 0) {
			if (diametre <= 0)
				diametre = 0.001;
			if (diametre2 <= 0)
				diametre2 = 0.001;
			if (distance <= 0)
				distance = 0.001;
	        
			imagep = calcul_imagep(pupille, diametre, diametre2, distance, cote);
			normalisation(imagep);
			double[][] figured = new double[dimx][dimy];
	        ListeValeurs out = new ListeValeurs();
			double[][] image_combinee = combinaison(imagep, figured);
	        out.ajouter("imagep", image_combinee);
	        out.ajouterString("info", "calcul de l'image de diffraction en cours...");
			etape = 1;
	        return out;
		} else if (etape == 1) {
		    double[][] figured;
		    if (zoom) {
    			double[][] imagepz = zoom_pupille(imagep);
    			figured = calcul_figured(imagepz);
    			normalisation(figured);
    			figured = zoom_figure(figured);
			} else {
    			figured = calcul_figured(imagep);
    			normalisation(figured);
			}
			
	        ListeValeurs out = new ListeValeurs();
			double[][] image_combinee = combinaison(imagep, figured);
	        out.ajouter("imagep", image_combinee);
	        //out.ajouterString("info", "calcul termin");
			double echelle = dimx * (lambda / cote) * (180 / Math.PI) * 3600;
	        out.ajouterString("info", "ct = " + joliFormattage(cote) + " m                                                          " +
				joliFormattage(echelle) + "''");
			etape = 2;
	        return out;
		} else
			return(null);
    }
    
    public ListeValeurs calcul(ListeValeurs in) throws SimuException {
        throw new SimuException("calcul n'est pas implment dans Pupilles: " +
            "utiliser calculLive  la place");
    }
    
	public static double[][] zoom_pupille(double[][] im) {
	// double la taille et copie l'image au centre, rempli de zros le reste
		int dimx = im[0].length;
		int dimy = im.length;
		double[][] res = new double[dimx*2][dimy*2];
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++)
				res[i][j] = 0;
		for (int i=dimx/2; i<(dimx*3)/2; i++)
			for (int j=dimy/2; j<(dimy*3)/2; j++)
				res[i][j] = im[i-dimx/2][j-dimy/2];
	    return(res);
	}
	
	public static double[][] zoom_figure(double[][] im) {
	// divise la taille d'un ct par 2, en gardant le centre de l'image
		int dimx = im[0].length;
		int dimy = im.length;
		double[][] res = new double[dimx/2][dimy/2];
		for (int i=0; i<dimx/2; i++)
			for (int j=0; j<dimy/2; j++)
				res[i][j] = im[dimx/4 + i][dimy/4 + j];
	    return(res);
	}
	
	public static double[][] combinaison(double[][] im1, double[][] im2) {
		int dx = im1.length + im2.length + 1;
		int dy = Math.max(im1[0].length, im2[0].length);
		double[][] im = new double[dx][dy];
		
		for (int i=0; i<im1.length; i++)
			for (int j=0; j<im1[0].length; j++)
				im[i][j] = im1[i][j];
		
		for (int j=0; j<im1[0].length; j++)
			im[im1.length][j] = 255;
		
		for (int i=0; i<im2.length; i++)
			for (int j=0; j<im2[0].length; j++)
				im[i + im1.length + 1][j] = im2[i][j];
		return(im);
	}
	
	public static void normalisation(double[][] t) {
		int dx = t.length;
		int dy = t[0].length;
		double max = 0;
		for (int i=0; i<dx; i++)
			for (int j=0; j<dy; j++)
				if (t[i][j] > max)
					max = t[i][j];
		for (int i=0; i<dx; i++)
			for (int j=0; j<dy; j++)
				if (t[i][j] < 0)
			 		t[i][j] = 0;
				else
					t[i][j] = t[i][j]*255/max;
	}
	
	/*public static void normalisation(double[][] t) {
		int dx = t.length;
		int dy = t[0].length;
		double moyenne = 0;
		double f=1/(1.0*dx*dy);
		for (int i=0; i<dx; i++)
			for (int j=0; j<dy; j++)
				moyenne += t[i][j]*f;
		System.out.println("moyenne="+moyenne);
		double v;
		for (int i=0; i<dx; i++)
			for (int j=0; j<dy; j++) {
				v = t[i][j]*255/(moyenne*2);
				if (v > 255)
					v = 255;
				t[i][j] = v;
			}
	}*/
	
	public static double[][] calcul_imagep(String pupille, double diametre, double diametre2, double distance, double cote) {
		double[][] imagep;
		
		if ("circulaire".equals(pupille))
			imagep = pupille_circulaire(diametre, cote);
		else if ("circ. + obst. second.".equals(pupille))
			imagep = pupille_obturation(diametre, diametre2, cote);
		else if ("2 circulaires".equals(pupille))
			imagep = pupille_2circ(diametre, distance, cote);
		else if ("croix d'Angel".equals(pupille))
			imagep = pupille_croix(diametre, distance, cote);
		else
			imagep = null;
		
		return(imagep);
	}
	
	public static double[][] pupille_circulaire(double diametre, double cote) {
		double[][] imagep = new double[dimx][dimy];
		int cx = dimx / 2;
		int cy = dimy / 2;
		//int r2 = Math.min(dimx, dimy) / 2;
		int r2 = (int)Math.round((diametre/2)*(dimx/cote));
		r2 = r2*r2;
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++) {
				if ((cx-i)*(cx-i) + (cy-j)*(cy-j) < r2)
					imagep[i][j] = 1;
				else
					imagep[i][j] = 0;
			}
		return(imagep);
	}
	
	public static double[][] pupille_obturation(double diametre, double diametre2, double cote) {
		double[][] imagep = new double[dimx][dimy];
		int cx = dimx / 2;
		int cy = dimy / 2;
		int r2 = (int)Math.round((diametre/2)*(dimx/cote));
		r2 = r2*r2;
		int r22 = (int)Math.round((diametre2/2)*(dimx/cote));
		r22 = r22*r22;
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++) {
				int rij = (cx-i)*(cx-i) + (cy-j)*(cy-j);
				if (rij < r2 && rij > r22)
					imagep[i][j] = 1;
				else
					imagep[i][j] = 0;
			}
		return(imagep);
	}
	
	public static double[][] pupille_2circ(double diametre, double distance, double cote) {
		double[][] imagep = new double[dimx][dimy];
		int distnorm = (int)Math.round((dimx/cote)*distance);
		int cx1 = (dimx - distnorm) / 2;
		int cx2 = (dimx + distnorm) / 2;
		int cy = dimy / 2;
		int r2 = (int)Math.round((diametre/2) * (dimx/cote));
		r2 = r2*r2;
		for (int i=0; i<dimx/2; i++)
			for (int j=0; j<dimy; j++) {
				if ((cx1-i)*(cx1-i) + (cy-j)*(cy-j) < r2)
					imagep[i][j] = 1;
				else
					imagep[i][j] = 0;
			}
		for (int i=dimx/2; i<dimx; i++)
			for (int j=0; j<dimy; j++) {
				if ((cx2-i)*(cx2-i) + (cy-j)*(cy-j) < r2)
					imagep[i][j] = 1;
				else
					imagep[i][j] = 0;
			}
		return(imagep);
	}
	
	public static void ajouterDisque(double[][] imagep, int cx, int cy, int r2) {
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++) {
				if ((cx-i)*(cx-i) + (cy-j)*(cy-j) < r2)
					imagep[i][j] = 1;
			}
	}
	
	public static double[][] pupille_croix(double diametre, double distance, double cote) {
		double[][] imagep = new double[dimx][dimy];
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++)
					imagep[i][j] = 0;
		int distnorm = (int)Math.round((dimx/cote)*distance);
		int r2 = (int)Math.round((diametre/2)*(dimx/cote));
		r2 = r2*r2;
		ajouterDisque(imagep, dimx / 2, (dimy - distnorm) / 2, r2);
		ajouterDisque(imagep, (dimx - distnorm) / 2, dimy / 2, r2);
		ajouterDisque(imagep, (dimx + distnorm) / 2, dimy / 2, r2);
		ajouterDisque(imagep, dimx / 2, (dimy + distnorm) / 2, r2);
		return(imagep);
	}
	
	public double[][] calcul_figured(double[][] imagep) {
		int dimx = imagep[0].length;
		int dimy = imagep.length;
		double[][] figured = new double[dimx][dimy];
		
		boolean log = ("logarithmique".equals(echelle));
		if ("FFT".equals(methode)) {
			TFLib.transfo2D_FFT(imagep, figured, log, quadrants);
		} else {
			double temp;
			for (int i=0; i<dimx; i++)
				for (int j=0; j<dimy; j++) {
					temp = transfo2D(imagep, i, j);
					if (log)
		            	figured[i][j] = Math.log(temp);
					else
		            	figured[i][j] = Math.sqrt(temp);
				}
		}
		return(figured);
	}
	
    public static double transfo2D(double[][] imagep, int x, int y) {
		double somme_r = 0.;
		double somme_i = 0.;
		double norm = (dimx + dimy) / 2;
		int cx = dimx / 2;
		int cy = dimy / 2;
		double p;
		double fact = 2 * Math.PI / norm;
		/*
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++) {
				p = (i-cx)*(x-cx) + (j-cy)*(y-cy); // attention, opration / entiers, marche avec petites valeurs
				somme += imagep[i][j] * Math.cos(p * fact);
			}
		*/
		// imagep[i][j] est forcment 0 ou 1 -> optimisation
		double xmcx = x - cx;
		double ymcy = y - cy;
		double imcx;
		double p1;
		for (int i=0; i<dimx; i++) {
			imcx = i - cx;
			p1 = imcx * xmcx;
			for (int j=0; j<dimy; j++)
				if (imagep[i][j] != 0) {
					p = (p1 + (j-cy)*ymcy) * fact;
					somme_r += Math.cos(p);
					somme_i += Math.sin(p);
				}
		}
		return(somme_r*somme_r + somme_i*somme_i);
	}
	
	// mthode d'affichage de double issue du Calcotron
    public static String joliFormattage(double v) {
        int precision = 3;
        int limitesci = 3; // dcimales avant de passer en mode sci
        
        if (v == 0)
            return("0");
        String sj = ""+v;
        if ("Infinity".equals(sj) || "-Infinity".equals(sj) || "NaN".equals(sj))
            return(""+v);
        String signe;
        if (v >= 0)
            signe = "";
        else {
            v = -v;
            signe = "-";
        }
        int exposant = (int)Math.floor(Math.log(v)/Math.log(10));
        double exparr = Math.pow(10, precision);
        double mantisse = v / Math.pow(10, exposant);
        mantisse = Math.rint(mantisse*exparr)/exparr;
        if (mantisse >= 10) {
            exposant += 1;
            mantisse = v / Math.pow(10, exposant);
            mantisse = Math.rint(mantisse*exparr)/exparr;
        }
        String sv = "";
        int chiffre = (int)Math.floor(mantisse);
        sv += chiffre + ".";
        for (int i=0; i<precision-1; i++) {
            mantisse = (mantisse - chiffre)*10;
            double exparr2 = Math.pow(10, precision-i);
            mantisse = Math.rint(mantisse*exparr)/exparr;
            chiffre = (int)Math.floor(mantisse);
            sv += chiffre;
        }
        mantisse = (mantisse - chiffre)*10;
        chiffre = (int)Math.round(mantisse);
        sv += chiffre;
        char dernier = sv.charAt(sv.length()-1);
        while (dernier == '0') {
            sv = sv.substring(0, sv.length()-1);
            dernier = sv.charAt(sv.length()-1);
        }
        if (exposant != 0) {
            if (exposant > 0 && exposant < limitesci) {
                String sd = sv.substring(0,1);
                if (sv.length()-2 < exposant) {
                    // transformation de 3.2E2 en 320
                    if (sv.length() > 2)
                        sd += sv.substring(2,sv.length());
                    for (int i=0; i<exposant-(sv.length()-2); i++)
                        sd += '0';
                } else {
                    // transformation de 3.245E2 en 324.5
                    if (sv.length() < exposant+2)
                        sd += sv.substring(2);
                    else
                        sd += sv.substring(2,exposant+2) + "." + sv.substring(exposant+2);
                }
                sv = sd;
            } else if (exposant < 0 && exposant > -limitesci) {
                // transformation de 3.2E-2 en 0.032
                String sd = "0.";
                for (int i=0; i<-exposant-1; i++)
                    sd += '0';
                sv = sd + sv.charAt(0) + sv.substring(2);
            } else
                sv += "E" + exposant;
        }
        if (sv.charAt(sv.length()-1) == '.')
            sv = sv.substring(0, sv.length()-1);
        
        sv = signe + sv;
        
        return(sv);
    }
    
}
