/*
Transformée de Fourier

Copyright (C) 2012 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU.

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/


function isCanvasSupported() {
    var elem = document.createElement('canvas');
    return !!(elem.getContext && elem.getContext('2d'));
}

function versionApplet() {
    var applet = document.createElement('applet');
    var width, height; // espace disponible pour l'applet
    if (typeof(window.innerWidth) == 'number') {
        // autre que IE
        width = window.innerWidth;
        height = window.innerHeight;
    } else if ( document.documentElement && (document.documentElement.clientWidth || document.documentElement.clientHeight) ) {
        // IE 6+ en mode 'standards compliant'
        width = document.documentElement.clientWidth;
        height = document.documentElement.clientHeight;
    } else {
        width = '100%';
        height = '100%';
    }
    applet.setAttribute('width', width);
    applet.setAttribute('height', height);
    applet.setAttribute('archive', 'Pupilles.jar,SimuLab.jar');
    applet.setAttribute('code', 'Pupilles');

    document.body.appendChild(applet);
}

if (isCanvasSupported()) {
    new Pupilles();
} else {
    // on ne peut pas utiliser excanvas ou flashcanvas pour manipuler des bitmaps offscreen
    // donc on utilise la version sous forme d'applet en alternative pour les vieux navigateurs
    if (window.addEventListener)
        window.addEventListener('load', function() { versionApplet(); }, false);
    else if (window.attachEvent)
        window.attachEvent('onload', function() { versionApplet(); });
}


function Pupilles() {
    this.applet = new SimuApplet(this);
    this.applet.appletDoc("pupilles_doc.xml");
    
	this.dimx = 256; // imgdimx dans le fichier doc.xml doit correspondre à 2*dimx + 1
	this.dimy = 256; // doit être égal à imgdimy
    this.pupille = '';
    this.diametre = 0.;
    this.diametre2 = 0.;
    this.distance = 0.;
	this.cote = 0.;
	this.methode = '';
	this.echelle = '';
	this.quadrants = true;
	this.lambda = 0.;
	this.zoom = false;
	this.imagep = null; // double[][]
	this.etape = 0;
}

Pupilles.prototype.initCalculLive = function(lin) {
    this.pupille = lin.lireString('pupille');
    this.diametre = lin.lireDouble('diametre');
    this.diametre2 = lin.lireDouble('diametre2');
    this.distance = lin.lireDouble('distance');
    this.cote = lin.lireDouble('cote');
    this.methode = lin.lireString('methode');
    this.echelle = lin.lireString('echelle');
    this.quadrants = (lin.lireString('quadrants') == 'on');
    this.lambda = lin.lireDouble('lambda');
    this.zoom = (lin.lireString('zoom') == 'on');
    this.etape = 0;
}

Pupilles.prototype.tableau2D = function(x, y) {
    var t, i, j;
    t = new Array(x);
    for (i=0; i<x; i++)
        t[i] = new Array(y);
    for (i=0; i<x; i++)
        for (j=0; j<y; j++)
            t[i][j] = 0;
    return(t);
}

Pupilles.prototype.calculLive = function() {
    var out, figured, image_combinee, imagepz, echelle;
    
    if (this.etape == 0) {
        if (this.diametre <= 0)
            this.diametre = 0.001;
        if (this.diametre2 <= 0)
            this.diametre2 = 0.001;
        if (this.distance <= 0)
            this.distance = 0.001;
        
        this.imagep = this.calcul_imagep(this.pupille, this.diametre, this.diametre2, this.distance, this.cote);
        this.normalisation(this.imagep);
        figured = this.tableau2D(this.dimx, this.dimy);
        out = new ListeValeurs();
        image_combinee = this.combinaison(this.imagep, figured);
        out.ajouter('imagep', image_combinee);
        out.ajouterString('info', "calcul de l'image de diffraction en cours...");
        this.etape = 1;
        return out;
        
    } else if (this.etape == 1) {
        if (this.zoom) {
            imagepz = this.zoom_pupille(this.imagep);
            figured = this.calcul_figured(imagepz);
            this.normalisation(figured);
            figured = this.zoom_figure(figured);
        } else {
            figured = this.calcul_figured(this.imagep);
            this.normalisation(figured);
        }
        
        out = new ListeValeurs();
        image_combinee = this.combinaison(this.imagep, figured);
        out.ajouter('imagep', image_combinee);
        //out.ajouterString("info", "calcul terminé");
        echelle = this.dimx * (this.lambda / this.cote) * (180 / Math.PI) * 3600;
        out.ajouterString('info', "image de gauche: côté = " + this.joliFormattage(this.cote, 3) + " m ; " +
            "image de droite: " + this.joliFormattage(echelle, 3) + "''");
        this.etape = 2;
        return out;
        
    } else
        return(null);
}

//public static double[][] zoom_pupille(double[][] im) {
Pupilles.prototype.zoom_pupille = function(im) {
// double la taille et copie l'image au centre, rempli de zéros le reste
    var dimx, dimy, res, i, j;
    dimx = im[0].length;
    dimy = im.length;
    res = this.tableau2D(dimx*2, dimy*2);
    for (i=0; i<dimx; i++)
        for (j=0; j<dimy; j++)
            res[i][j] = 0;
    for (i=dimx/2; i<(dimx*3)/2; i++)
        for (j=dimy/2; j<(dimy*3)/2; j++)
            res[i][j] = im[i-dimx/2][j-dimy/2];
    return(res);
}

//public static double[][] zoom_figure(double[][] im) {
Pupilles.prototype.zoom_figure = function(im) {
// divise la taille d'un côté par 2, en gardant le centre de l'image
    var dimx, dimy, res, i, j;
    dimx = im[0].length;
    dimy = im.length;
    res = this.tableau2D(dimx/2, dimy/2);
    for (i=0; i<dimx/2; i++)
        for (j=0; j<dimy/2; j++)
            res[i][j] = im[dimx/4 + i][dimy/4 + j];
    return(res);
}

//public static double[][] combinaison(double[][] im1, double[][] im2) {
Pupilles.prototype.combinaison = function(im1, im2) {
    var dx, dy, im, i, j;
    dx = im1.length + im2.length + 1;
    dy = Math.max(im1[0].length, im2[0].length);
    im = this.tableau2D(dx, dy);
    
    for (i=0; i<im1.length; i++)
        for (j=0; j<im1[0].length; j++)
            im[i][j] = im1[i][j];
    
    for (j=0; j<im1[0].length; j++)
        im[im1.length][j] = 255;
    
    for (i=0; i<im2.length; i++)
        for (j=0; j<im2[0].length; j++)
            im[i + im1.length + 1][j] = im2[i][j];
    return(im);
}

//public static void normalisation(double[][] t) {
Pupilles.prototype.normalisation = function(t) {
    var dx, dy, max, i, j;
    dx = t.length;
    dy = t[0].length;
    max = 0;
    for (i=0; i<dx; i++)
        for (j=0; j<dy; j++)
            if (t[i][j] > max)
                max = t[i][j];
    for (i=0; i<dx; i++)
        for (j=0; j<dy; j++)
            if (t[i][j] < 0)
                t[i][j] = 0;
            else
                t[i][j] = t[i][j]*255/max;
}

//public static double[][] calcul_imagep(String pupille, double diametre, double diametre2, double distance, double cote) {
Pupilles.prototype.calcul_imagep = function(pupille, diametre, diametre2, distance, cote) {
    var imagep;
    
    if (pupille == "circulaire")
        imagep = this.pupille_circulaire(diametre, cote);
    else if (pupille == "circ. + obst. second.")
        imagep = this.pupille_obturation(diametre, diametre2, cote);
    else if (pupille == "2 circulaires")
        imagep = this.pupille_2circ(diametre, distance, cote);
    else if (pupille == "croix d'Angel")
        imagep = this.pupille_croix(diametre, distance, cote);
    else
        imagep = null;
    
    return(imagep);
}

//public static double[][] pupille_circulaire(double diametre, double cote) {
Pupilles.prototype.pupille_circulaire = function(diametre, cote) {
    var imagep, cx, cy, r2, i, j;
    imagep = this.tableau2D(this.dimx, this.dimy);
    cx = this.dimx / 2;
    cy = this.dimy / 2;
    r2 = Math.round((diametre/2)*(this.dimx/cote));
    r2 = r2*r2;
    for (i=0; i<this.dimx; i++)
        for (j=0; j<this.dimy; j++) {
            if ((cx-i)*(cx-i) + (cy-j)*(cy-j) < r2)
                imagep[i][j] = 1;
            else
                imagep[i][j] = 0;
        }
    return(imagep);
}

//public static double[][] pupille_obturation(double diametre, double diametre2, double cote) {
Pupilles.prototype.pupille_obturation = function(diametre, diametre2, cote) {
    var imagep, cx, cy, r2, r22, i, j, rij;
    imagep = this.tableau2D(this.dimx, this.dimy);
    cx = this.dimx / 2;
    cy = this.dimy / 2;
    r2 = Math.round((diametre/2)*(this.dimx/cote));
    r2 = r2*r2;
    r22 = Math.round((diametre2/2)*(this.dimx/cote));
    r22 = r22*r22;
    for (i=0; i<this.dimx; i++)
        for (j=0; j<this.dimy; j++) {
            rij = (cx-i)*(cx-i) + (cy-j)*(cy-j);
            if (rij < r2 && rij > r22)
                imagep[i][j] = 1;
            else
                imagep[i][j] = 0;
        }
    return(imagep);
}

//public static double[][] pupille_2circ(double diametre, double distance, double cote) {
Pupilles.prototype.pupille_2circ = function(diametre, distance, cote) {
    var imagep, distnorm, cx1, cx2, cy, r2, i, j;
    imagep = this.tableau2D(this.dimx, this.dimy);
    distnorm = Math.round((this.dimx/cote)*distance);
    cx1 = (this.dimx - distnorm) / 2;
    cx2 = (this.dimx + distnorm) / 2;
    cy = this.dimy / 2;
    r2 = Math.round((diametre/2) * (this.dimx/cote));
    r2 = r2*r2;
    for (i=0; i<this.dimx/2; i++)
        for (j=0; j<this.dimy; j++) {
            if ((cx1-i)*(cx1-i) + (cy-j)*(cy-j) < r2)
                imagep[i][j] = 1;
            else
                imagep[i][j] = 0;
        }
    for (i=this.dimx/2; i<this.dimx; i++)
        for (j=0; j<this.dimy; j++) {
            if ((cx2-i)*(cx2-i) + (cy-j)*(cy-j) < r2)
                imagep[i][j] = 1;
            else
                imagep[i][j] = 0;
        }
    return(imagep);
}

//public static void ajouterDisque(double[][] imagep, int cx, int cy, int r2) {
Pupilles.prototype.ajouterDisque = function(imagep, cx, cy, r2) {
    var i, j;
    for (i=0; i<this.dimx; i++)
        for (j=0; j<this.dimy; j++) {
            if ((cx-i)*(cx-i) + (cy-j)*(cy-j) < r2)
                imagep[i][j] = 1;
        }
}

//public static double[][] pupille_croix(double diametre, double distance, double cote) {
Pupilles.prototype.pupille_croix = function(diametre, distance, cote) {
    var imagep, i, j, distnorm, r2;
    imagep = this.tableau2D(this.dimx, this.dimy);
    for (i=0; i<this.dimx; i++)
        for (j=0; j<this.dimy; j++)
                imagep[i][j] = 0;
    distnorm = Math.round((this.dimx/cote)*distance);
    r2 = Math.round((diametre/2)*(this.dimx/cote));
    r2 = r2*r2;
    this.ajouterDisque(imagep, this.dimx / 2, (this.dimy - distnorm) / 2, r2);
    this.ajouterDisque(imagep, (this.dimx - distnorm) / 2, this.dimy / 2, r2);
    this.ajouterDisque(imagep, (this.dimx + distnorm) / 2, this.dimy / 2, r2);
    this.ajouterDisque(imagep, this.dimx / 2, (this.dimy + distnorm) / 2, r2);
    return(imagep);
}

//public double[][] calcul_figured(double[][] imagep) {
Pupilles.prototype.calcul_figured = function(imagep) {
    var dimx, dimy, figured, log, temp, i, j;
    dimx = imagep[0].length;
    dimy = imagep.length;
    figured = this.tableau2D(dimx, dimy);
    
    log = (this.echelle == "logarithmique");
    if (this.methode == "FFT") {
        TFLib.transfo2D_FFT(imagep, figured, log, this.quadrants);
    } else {
        for (i=0; i<this.dimx; i++)
            for (j=0; j<this.dimy; j++) {
                temp = this.transfo2D(imagep, i, j);
                if (log)
                    figured[i][j] = Math.log(temp);
                else
                    figured[i][j] = Math.sqrt(temp);
            }
    }
    return(figured);
}

//public static double transfo2D(double[][] imagep, int x, int y) {
Pupilles.prototype.transfo2D = function(imagep, x, y) {
    var somme_r, somme_i, norm, cx, cy, p, fact, xmcx, ymcy, imcx, p1, i, j;
    somme_r = 0.;
    somme_i = 0.;
    norm = (this.dimx + this.dimy) / 2;
    cx = this.dimx / 2;
    cy = this.dimy / 2;
    this.fact = 2 * Math.PI / norm;
    /*
    for (int i=0; i<dimx; i++)
        for (int j=0; j<dimy; j++) {
            p = (i-cx)*(x-cx) + (j-cy)*(y-cy); // attention, opération / entiers, marche avec petites valeurs
            somme += imagep[i][j] * Math.cos(p * fact);
        }
    */
    // imagep[i][j] est forcément 0 ou 1 -> optimisation
    xmcx = x - cx;
    ymcy = y - cy;
    for (i=0; i<this.dimx; i++) {
        imcx = i - cx;
        p1 = imcx * xmcx;
        for (j=0; j<this.dimy; j++)
            if (imagep[i][j] != 0) {
                p = (p1 + (j-cy)*ymcy) * fact;
                somme_r += Math.cos(p);
                somme_i += Math.sin(p);
            }
    }
    return(somme_r*somme_r + somme_i*somme_i);
}

Pupilles.prototype.joliFormattage = function(d, precision) {
    if (isNaN(d))
        return('');
    if (d == 0)
        return('0');
    var facteur_precision = Math.pow(10, precision);
    var exposant = Math.floor(Math.log(Math.abs(d))/Math.log(10));
    var mantisse = d / (Math.pow(10,exposant));
    var s;
    if (exposant == 0) {
        mantisse = Math.round(mantisse*facteur_precision)/facteur_precision;
        s = ""+mantisse;
    } else if (exposant == 1) {
        mantisse = Math.round(mantisse*facteur_precision)/(facteur_precision/10);
        s = ""+mantisse;
    } else if (exposant == -1) {
        var signe;
        if (mantisse < 0)
            signe = "-";
        else
            signe = "";
        mantisse = Math.abs(Math.round(mantisse*facteur_precision));
        s = signe + "0."+Math.floor(mantisse);
        s = s.replace(/0+$/g, '');
    } else {
        mantisse = Math.round(mantisse*facteur_precision)/facteur_precision;
        s = mantisse + "E" + exposant;
    }
    return(s);
}
